@@ -2,7 +2,7 @@
 
   class.c -
 
-  $Author: nobu $
+  $Author$
   created at: Tue Aug 10 15:05:44 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -32,6 +32,9 @@
 
 #define id_attached id__attached__
 
+#define METACLASS_OF(k) RBASIC(k)->klass
+#define SET_METACLASS_OF(k, cls) RBASIC_SET_CLASS(k, cls)
+
 void
 rb_class_subclass_add(VALUE super, VALUE klass)
 {
@@ -176,7 +179,7 @@ class_alloc(VALUE flags, VALUE klass)
      */
     RCLASS_SET_ORIGIN((VALUE)obj, (VALUE)obj);
     RCLASS_SERIAL(obj) = rb_next_class_serial();
-    RCLASS_REFINED_CLASS(obj) = Qnil;
+    RB_OBJ_WRITE(obj, &RCLASS_REFINED_CLASS(obj), Qnil);
     RCLASS_EXT(obj)->allocator = 0;
 
     return (VALUE)obj;
@@ -205,7 +208,6 @@ rb_class_boot(VALUE super)
     RCLASS_SET_SUPER(klass, super);
     RCLASS_M_TBL_INIT(klass);
 
-    OBJ_INFECT(klass, super);
     return (VALUE)klass;
 }
 
@@ -313,6 +315,12 @@ class_init_copy_check(VALUE clone, VALUE orig)
 VALUE
 rb_mod_init_copy(VALUE clone, VALUE orig)
 {
+    /* cloned flag is refer at constant inline cache
+     * see vm_get_const_key_cref() in vm_insnhelper.c
+     */
+    FL_SET(clone, RCLASS_CLONED);
+    FL_SET(orig , RCLASS_CLONED);
+
     if (RB_TYPE_P(clone, T_CLASS)) {
 	class_init_copy_check(clone, orig);
     }
@@ -335,7 +343,7 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
     if (RCLASS_IV_TBL(orig)) {
 	st_data_t id;
 
-	RCLASS_IV_TBL(clone) = rb_st_copy(clone, RCLASS_IV_TBL(orig));
+	rb_iv_tbl_copy(clone, orig);
 	CONST_ID(id, "__tmp_classpath__");
 	st_delete(RCLASS_IV_TBL(clone), &id, 0);
 	CONST_ID(id, "__classpath__");
@@ -367,28 +375,41 @@ rb_singleton_class_clone(VALUE obj)
     return rb_singleton_class_clone_and_attach(obj, Qundef);
 }
 
+// Clone and return the singleton class of `obj` if it has been created and is attached to `obj`.
 VALUE
 rb_singleton_class_clone_and_attach(VALUE obj, VALUE attach)
 {
     const VALUE klass = RBASIC(obj)->klass;
 
-    if (!FL_TEST(klass, FL_SINGLETON))
-	return klass;
+    // Note that `rb_singleton_class()` can create situations where `klass` is
+    // attached to an object other than `obj`. In which case `obj` does not have
+    // a material singleton class attached yet and there is no singleton class
+    // to clone.
+    if (!(FL_TEST(klass, FL_SINGLETON) && rb_attr_get(klass, id_attached) == obj)) {
+        // nothing to clone
+        return klass;
+    }
     else {
 	/* copy singleton(unnamed) class */
+        bool klass_of_clone_is_new;
 	VALUE clone = class_alloc(RBASIC(klass)->flags, 0);
 
 	if (BUILTIN_TYPE(obj) == T_CLASS) {
+            klass_of_clone_is_new = true;
 	    RBASIC_SET_CLASS(clone, clone);
 	}
 	else {
-	    RBASIC_SET_CLASS(clone, rb_singleton_class_clone(klass));
+            VALUE klass_metaclass_clone = rb_singleton_class_clone(klass);
+            // When `METACLASS_OF(klass) == klass_metaclass_clone`, it means the
+            // recursive call did not clone `METACLASS_OF(klass)`.
+            klass_of_clone_is_new = (METACLASS_OF(klass) != klass_metaclass_clone);
+            RBASIC_SET_CLASS(clone, klass_metaclass_clone);
 	}
 
 	RCLASS_SET_SUPER(clone, RCLASS_SUPER(klass));
 	RCLASS_EXT(clone)->allocator = RCLASS_EXT(klass)->allocator;
 	if (RCLASS_IV_TBL(klass)) {
-	    RCLASS_IV_TBL(clone) = rb_st_copy(clone, RCLASS_IV_TBL(klass));
+	    rb_iv_tbl_copy(clone, klass);
 	}
 	if (RCLASS_CONST_TBL(klass)) {
 	    struct clone_const_arg arg;
@@ -406,7 +427,9 @@ rb_singleton_class_clone_and_attach(VALUE obj, VALUE attach)
 	    arg.new_klass = clone;
 	    rb_id_table_foreach(RCLASS_M_TBL(klass), clone_method_i, &arg);
 	}
-	rb_singleton_class_attached(RBASIC(clone)->klass, clone);
+        if (klass_of_clone_is_new) {
+            rb_singleton_class_attached(RBASIC(clone)->klass, clone);
+        }
 	FL_SET(clone, FL_SINGLETON);
 
 	return clone;
@@ -428,11 +451,6 @@ rb_singleton_class_attached(VALUE klass, VALUE obj)
     }
 }
 
-
-
-#define METACLASS_OF(k) RBASIC(k)->klass
-#define SET_METACLASS_OF(k, cls) RBASIC_SET_CLASS(k, cls)
-
 /*!
  * whether k is a meta^(n)-class of Class class
  * @retval 1 if \a k is a meta^(n)-class of Class class (n >= 0)
@@ -505,8 +523,6 @@ make_metaclass(VALUE klass)
     while (RB_TYPE_P(super, T_ICLASS)) super = RCLASS_SUPER(super);
     RCLASS_SET_SUPER(metaclass, super ? ENSURE_EIGENCLASS(super) : rb_cClass);
 
-    OBJ_INFECT(metaclass, RCLASS_SUPER(metaclass));
-
     return metaclass;
 }
 
@@ -537,8 +553,8 @@ boot_defclass(const char *name, VALUE super)
     VALUE obj = rb_class_boot(super);
     ID id = rb_intern(name);
 
-    rb_name_class(obj, id);
     rb_const_set((rb_cObject ? rb_cObject : obj), id, obj);
+    rb_vm_add_root_module(id, obj);
     return obj;
 }
 
@@ -550,7 +566,7 @@ Init_class_hierarchy(void)
     rb_gc_register_mark_object(rb_cObject);
 
     /* resolve class name ASAP for order-independence */
-    rb_class_name(rb_cObject);
+    rb_set_class_path_string(rb_cObject, rb_cObject, rb_fstring_lit("Object"));
 
     rb_cModule = boot_defclass("Module", rb_cObject);
     rb_cClass =  boot_defclass("Class",  rb_cModule);
@@ -658,6 +674,9 @@ rb_define_class(const char *name, VALUE super)
 	if (rb_class_real(RCLASS_SUPER(klass)) != super) {
 	    rb_raise(rb_eTypeError, "superclass mismatch for class %s", name);
 	}
+
+        /* Class may have been defined in Ruby and not pin-rooted */
+        rb_vm_add_root_module(id, klass);
 	return klass;
     }
     if (!super) {
@@ -665,7 +684,6 @@ rb_define_class(const char *name, VALUE super)
     }
     klass = rb_define_class_id(id, super);
     rb_vm_add_root_module(id, klass);
-    rb_name_class(klass, id);
     rb_const_set(rb_cObject, id, klass);
     rb_class_inherited(super, klass);
 
@@ -730,6 +748,9 @@ rb_define_class_id_under(VALUE outer, ID id, VALUE super)
 		     " (%"PRIsVALUE" is given but was %"PRIsVALUE")",
 		     outer, rb_id2str(id), RCLASS_SUPER(klass), super);
 	}
+        /* Class may have been defined in Ruby and not pin-rooted */
+        rb_vm_add_root_module(id, klass);
+
 	return klass;
     }
     if (!super) {
@@ -740,6 +761,7 @@ rb_define_class_id_under(VALUE outer, ID id, VALUE super)
     rb_set_class_path_string(klass, outer, rb_id2str(id));
     rb_const_set(outer, id, klass);
     rb_class_inherited(super, klass);
+    rb_vm_add_root_module(id, klass);
     rb_gc_register_mark_object(klass);
 
     return klass;
@@ -756,12 +778,7 @@ rb_module_new(void)
 VALUE
 rb_define_module_id(ID id)
 {
-    VALUE mdl;
-
-    mdl = rb_module_new();
-    rb_name_class(mdl, id);
-
-    return mdl;
+    return rb_module_new();
 }
 
 VALUE
@@ -777,10 +794,13 @@ rb_define_module(const char *name)
 	    rb_raise(rb_eTypeError, "%s is not a module (%"PRIsVALUE")",
 		     name, rb_obj_class(module));
 	}
+        /* Module may have been defined in Ruby and not pin-rooted */
+        rb_vm_add_root_module(id, module);
 	return module;
     }
     module = rb_define_module_id(id);
     rb_vm_add_root_module(id, module);
+    rb_gc_register_mark_object(module);
     rb_const_set(rb_cObject, id, module);
 
     return module;
@@ -819,6 +839,10 @@ rb_include_class_new(VALUE module, VALUE super)
 {
     VALUE klass = class_alloc(T_ICLASS, rb_cClass);
 
+    RCLASS_M_TBL(OBJ_WB_UNPROTECT(klass)) =
+      RCLASS_M_TBL(OBJ_WB_UNPROTECT(module)); /* TODO: unprotected? */
+
+    RCLASS_SET_ORIGIN(klass, module == RCLASS_ORIGIN(module) ? klass : RCLASS_ORIGIN(module));
     if (BUILTIN_TYPE(module) == T_ICLASS) {
 	module = RBASIC(module)->klass;
     }
@@ -831,9 +855,6 @@ rb_include_class_new(VALUE module, VALUE super)
     RCLASS_IV_TBL(klass) = RCLASS_IV_TBL(module);
     RCLASS_CONST_TBL(klass) = RCLASS_CONST_TBL(module);
 
-    RCLASS_M_TBL(OBJ_WB_UNPROTECT(klass)) =
-      RCLASS_M_TBL(OBJ_WB_UNPROTECT(RCLASS_ORIGIN(module))); /* TODO: unprotected? */
-
     RCLASS_SET_SUPER(klass, super);
     if (RB_TYPE_P(module, T_ICLASS)) {
 	RBASIC_SET_CLASS(klass, RBASIC(module)->klass);
@@ -841,8 +862,6 @@ rb_include_class_new(VALUE module, VALUE super)
     else {
 	RBASIC_SET_CLASS(klass, module);
     }
-    OBJ_INFECT(klass, module);
-    OBJ_INFECT(klass, super);
 
     return (VALUE)klass;
 }
@@ -857,7 +876,6 @@ ensure_includable(VALUE klass, VALUE module)
     if (!NIL_P(rb_refinement_module_get_refined_class(module))) {
 	rb_raise(rb_eArgError, "refinement module is not allowed");
     }
-    OBJ_INFECT(klass, module);
 }
 
 void
@@ -879,6 +897,8 @@ add_refined_method_entry_i(ID key, VALUE value, void *data)
     return ID_TABLE_CONTINUE;
 }
 
+static void ensure_origin(VALUE klass);
+
 static int
 include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super)
 {
@@ -886,20 +906,27 @@ include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super)
     int method_changed = 0, constant_changed = 0;
     struct rb_id_table *const klass_m_tbl = RCLASS_M_TBL(RCLASS_ORIGIN(klass));
 
+    if (FL_TEST(module, RCLASS_REFINED_BY_ANY)) {
+        ensure_origin(module);
+    }
+
     while (module) {
+        int origin_seen = FALSE;
 	int superclass_seen = FALSE;
 	struct rb_id_table *tbl;
 
-	if (RCLASS_ORIGIN(module) != module)
-	    goto skip;
+        if (klass == c)
+            origin_seen = TRUE;
 	if (klass_m_tbl && klass_m_tbl == RCLASS_M_TBL(module))
 	    return -1;
 	/* ignore if the module included already in superclasses */
 	for (p = RCLASS_SUPER(klass); p; p = RCLASS_SUPER(p)) {
 	    int type = BUILTIN_TYPE(p);
+            if (c == p)
+                origin_seen = TRUE;
 	    if (type == T_ICLASS) {
 		if (RCLASS_M_TBL(p) == RCLASS_M_TBL(module)) {
-		    if (!superclass_seen) {
+		    if (!superclass_seen && origin_seen) {
 			c = p;  /* move insertion point */
 		    }
 		    goto skip;
@@ -912,6 +939,7 @@ include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super)
 	}
 	iclass = rb_include_class_new(module, RCLASS_SUPER(c));
 	c = RCLASS_SET_SUPER(c, iclass);
+        RCLASS_SET_INCLUDER(iclass, klass);
 
 	{
 	    VALUE m = module;
@@ -969,15 +997,10 @@ move_refined_method(ID key, VALUE value, void *data)
     }
 }
 
-void
-rb_prepend_module(VALUE klass, VALUE module)
+static void
+ensure_origin(VALUE klass)
 {
-    VALUE origin;
-    int changed = 0;
-
-    ensure_includable(klass, module);
-
-    origin = RCLASS_ORIGIN(klass);
+    VALUE origin = RCLASS_ORIGIN(klass);
     if (origin == klass) {
 	origin = class_alloc(T_ICLASS, klass);
 	OBJ_WB_UNPROTECT(origin); /* TODO: conservative shading. Need more survey. */
@@ -988,6 +1011,15 @@ rb_prepend_module(VALUE klass, VALUE module)
 	RCLASS_M_TBL_INIT(klass);
 	rb_id_table_foreach(RCLASS_M_TBL(origin), move_refined_method, (void *)klass);
     }
+}
+
+void
+rb_prepend_module(VALUE klass, VALUE module)
+{
+    int changed = 0;
+
+    ensure_includable(klass, module);
+    ensure_origin(klass);
     changed = include_modules_at(klass, klass, module, FALSE);
     if (changed < 0)
 	rb_raise(rb_eArgError, "cyclic prepend detected");
@@ -1087,10 +1119,11 @@ rb_mod_ancestors(VALUE mod)
     VALUE p, ary = rb_ary_new();
 
     for (p = mod; p; p = RCLASS_SUPER(p)) {
+        if (p != RCLASS_ORIGIN(p)) continue;
 	if (BUILTIN_TYPE(p) == T_ICLASS) {
 	    rb_ary_push(ary, RBASIC(p)->klass);
 	}
-	else if (p == RCLASS_ORIGIN(p)) {
+        else {
 	    rb_ary_push(ary, p);
 	}
     }
@@ -1162,7 +1195,7 @@ method_entry_i(ID key, VALUE value, void *data)
 	if (!me) return ID_TABLE_CONTINUE;
 	if (!arg->recur && me->owner != owner) return ID_TABLE_CONTINUE;
     }
-    if (!st_lookup(arg->list, key, 0)) {
+    if (!st_is_member(arg->list, key)) {
 	if (UNDEFINED_METHOD_ENTRY_P(me)) {
 	    type = METHOD_VISI_UNDEF; /* none */
 	}
@@ -1174,6 +1207,23 @@ method_entry_i(ID key, VALUE value, void *data)
     return ID_TABLE_CONTINUE;
 }
 
+static void
+add_instance_method_list(VALUE mod, struct method_entry_arg *me_arg)
+{
+    struct rb_id_table *m_tbl = RCLASS_M_TBL(mod);
+    if (!m_tbl) return;
+    rb_id_table_foreach(m_tbl, method_entry_i, me_arg);
+}
+
+static bool
+particular_class_p(VALUE mod)
+{
+    if (!mod) return false;
+    if (FL_TEST(mod, FL_SINGLETON)) return true;
+    if (BUILTIN_TYPE(mod) == T_ICLASS) return true;
+    return false;
+}
+
 static VALUE
 class_instance_method_list(int argc, const VALUE *argv, VALUE mod, int obj, int (*func) (st_data_t, st_data_t, st_data_t))
 {
@@ -1183,20 +1233,26 @@ class_instance_method_list(int argc, const VALUE *argv, VALUE mod, int obj, int
 
     if (rb_check_arity(argc, 0, 1)) recur = RTEST(argv[0]);
 
+    me_arg.list = st_init_numtable();
+    me_arg.recur = recur;
+
+    if (obj) {
+        for (; particular_class_p(mod); mod = RCLASS_SUPER(mod)) {
+            add_instance_method_list(mod, &me_arg);
+        }
+    }
+
     if (!recur && RCLASS_ORIGIN(mod) != mod) {
 	mod = RCLASS_ORIGIN(mod);
 	prepended = 1;
     }
 
-    me_arg.list = st_init_numtable();
-    me_arg.recur = recur;
     for (; mod; mod = RCLASS_SUPER(mod)) {
-	if (RCLASS_M_TBL(mod)) rb_id_table_foreach(RCLASS_M_TBL(mod), method_entry_i, &me_arg);
+        add_instance_method_list(mod, &me_arg);
 	if (BUILTIN_TYPE(mod) == T_ICLASS && !prepended) continue;
-	if (obj && FL_TEST(mod, FL_SINGLETON)) continue;
 	if (!recur) break;
     }
-    ary = rb_ary_new();
+    ary = rb_ary_new2(me_arg.list->num_entries);
     st_foreach(me_arg.list, func, ary);
     st_free_table(me_arg.list);
 
@@ -1297,7 +1353,7 @@ rb_class_public_instance_methods(int argc, const VALUE *argv, VALUE mod)
  *  <i>obj</i>. This will include all the methods accessible in
  *  <i>obj</i>'s ancestors.
  *  If the optional parameter is <code>false</code>, it
- *  returns an array of <i>obj<i>'s public and protected singleton methods,
+ *  returns an array of <i>obj</i>'s public and protected singleton methods,
  *  the array will not include methods in modules included in <i>obj</i>.
  *
  *     class Klass
@@ -1416,6 +1472,9 @@ rb_obj_singleton_methods(int argc, const VALUE *argv, VALUE obj)
     int recur = TRUE;
 
     if (rb_check_arity(argc, 0, 1)) recur = RTEST(argv[0]);
+    if (RB_TYPE_P(obj, T_CLASS) && FL_TEST(obj, FL_SINGLETON)) {
+        rb_singleton_class(obj);
+    }
     klass = CLASS_OF(obj);
     origin = RCLASS_ORIGIN(klass);
     me_arg.list = st_init_numtable();
@@ -1430,7 +1489,7 @@ rb_obj_singleton_methods(int argc, const VALUE *argv, VALUE obj)
 	    klass = RCLASS_SUPER(klass);
 	}
     }
-    ary = rb_ary_new();
+    ary = rb_ary_new2(me_arg.list->num_entries);
     st_foreach(me_arg.list, ins_methods_i, ary);
     st_free_table(me_arg.list);
 
@@ -1494,24 +1553,36 @@ rb_obj_singleton_methods(int argc, const VALUE *argv, VALUE obj)
  * \{
  */
 
+#ifdef rb_define_method_id
+#undef rb_define_method_id
+#endif
 void
 rb_define_method_id(VALUE klass, ID mid, VALUE (*func)(ANYARGS), int argc)
 {
     rb_add_method_cfunc(klass, mid, func, argc, METHOD_VISI_PUBLIC);
 }
 
+#ifdef rb_define_method
+#undef rb_define_method
+#endif
 void
 rb_define_method(VALUE klass, const char *name, VALUE (*func)(ANYARGS), int argc)
 {
     rb_add_method_cfunc(klass, rb_intern(name), func, argc, METHOD_VISI_PUBLIC);
 }
 
+#ifdef rb_define_protected_method
+#undef rb_define_protected_method
+#endif
 void
 rb_define_protected_method(VALUE klass, const char *name, VALUE (*func)(ANYARGS), int argc)
 {
     rb_add_method_cfunc(klass, rb_intern(name), func, argc, METHOD_VISI_PROTECTED);
 }
 
+#ifdef rb_define_private_method
+#undef rb_define_private_method
+#endif
 void
 rb_define_private_method(VALUE klass, const char *name, VALUE (*func)(ANYARGS), int argc)
 {
@@ -1612,12 +1683,6 @@ singleton_class_of(VALUE obj)
 	RCLASS_SERIAL(klass) = serial;
     }
 
-    if (OBJ_TAINTED(obj)) {
-	OBJ_TAINT(klass);
-    }
-    else {
-	FL_UNSET(klass, FL_TAINT);
-    }
     RB_FL_SET_RAW(klass, RB_OBJ_FROZEN_RAW(obj));
 
     return klass;
@@ -1694,6 +1759,9 @@ rb_singleton_class(VALUE obj)
  * \{
  */
 
+#ifdef rb_define_singleton_method
+#undef rb_define_singleton_method
+#endif
 /*!
  * Defines a singleton method for \a obj.
  * \param obj    an arbitrary object
@@ -1707,8 +1775,9 @@ rb_define_singleton_method(VALUE obj, const char *name, VALUE (*func)(ANYARGS),
     rb_define_method(singleton_class_of(obj), name, func, argc);
 }
 
-
-
+#ifdef rb_define_module_function
+#undef rb_define_module_function
+#endif
 /*!
  * Defines a module function for \a module.
  * \param module  an module or a class.
@@ -1723,7 +1792,9 @@ rb_define_module_function(VALUE module, const char *name, VALUE (*func)(ANYARGS)
     rb_define_singleton_method(module, name, func, argc);
 }
 
-
+#ifdef rb_define_global_function
+#undef rb_define_global_function
+#endif
 /*!
  * Defines a global function
  * \param name    name of the function
@@ -1772,8 +1843,7 @@ rb_keyword_error_new(const char *error, VALUE keys)
 	rb_str_cat_cstr(error_message, ": ");
 	while (1) {
             const VALUE k = RARRAY_AREF(keys, i);
-	    Check_Type(k, T_SYMBOL); /* wrong hash is given to rb_get_kwargs */
-	    rb_str_append(error_message, rb_sym2str(k));
+	    rb_str_append(error_message, rb_inspect(k));
 	    if (++i >= len) break;
 	    rb_str_cat_cstr(error_message, ", ");
 	}
@@ -1801,57 +1871,33 @@ unknown_keyword_error(VALUE hash, const ID *table, int keywords)
     rb_keyword_error("unknown", rb_hash_keys(hash));
 }
 
-struct extract_keywords {
-    VALUE kwdhash, nonsymkey;
-};
 
 static int
 separate_symbol(st_data_t key, st_data_t value, st_data_t arg)
 {
-    struct extract_keywords *argp = (struct extract_keywords *)arg;
-    VALUE k = (VALUE)key, v = (VALUE)value;
-
-    if (argp->kwdhash) {
-        if (UNLIKELY(!SYMBOL_P(k))) {
-            argp->nonsymkey = k;
-            return ST_STOP;
-        }
-    }
-    else if (SYMBOL_P(k)) {
-        if (UNLIKELY(argp->nonsymkey != Qundef)) {
-            argp->kwdhash = Qnil;
-            return ST_STOP;
-        }
-        argp->kwdhash = rb_hash_new();
-    }
-    else {
-        if (argp->nonsymkey == Qundef)
-            argp->nonsymkey = k;
-        return ST_CONTINUE;
-    }
-    rb_hash_aset(argp->kwdhash, k, v);
+    VALUE *kwdhash = (VALUE *)arg;
+    if (!SYMBOL_P(key)) kwdhash++;
+    if (!*kwdhash) *kwdhash = rb_hash_new();
+    rb_hash_aset(*kwdhash, (VALUE)key, (VALUE)value);
     return ST_CONTINUE;
 }
 
 VALUE
 rb_extract_keywords(VALUE *orighash)
 {
-    struct extract_keywords arg = {0, Qundef};
+    VALUE parthash[2] = {0, 0};
     VALUE hash = *orighash;
 
     if (RHASH_EMPTY_P(hash)) {
 	*orighash = 0;
 	return hash;
     }
-    rb_hash_foreach(hash, separate_symbol, (st_data_t)&arg);
-    if (arg.kwdhash) {
-        if (arg.nonsymkey != Qundef) {
-            rb_raise(rb_eArgError, "non-symbol key in keyword arguments: %+"PRIsVALUE,
-                     arg.nonsymkey);
-        }
-        *orighash = 0;
+    rb_hash_foreach(hash, separate_symbol, (st_data_t)&parthash);
+    *orighash = parthash[1];
+    if (parthash[1] && RBASIC_CLASS(hash) != rb_cHash) {
+        RBASIC_SET_CLASS(parthash[1], RBASIC_CLASS(hash));
     }
-    return arg.kwdhash;
+    return parthash[0];
 }
 
 int
@@ -1864,8 +1910,8 @@ rb_get_kwargs(VALUE keyword_hash, const ID *table, int required, int optional, V
 
 #define extract_kwarg(keyword, val) \
     (key = (st_data_t)(keyword), values ? \
-     rb_hash_stlike_delete(keyword_hash, &key, (val)) : \
-     rb_hash_stlike_lookup(keyword_hash, key, (val)))
+     (rb_hash_stlike_delete(keyword_hash, &key, &(val)) || ((val) = Qundef, 0)) : \
+     rb_hash_stlike_lookup(keyword_hash, key, NULL))
 
     if (NIL_P(keyword_hash)) keyword_hash = 0;
 
@@ -1873,18 +1919,11 @@ rb_get_kwargs(VALUE keyword_hash, const ID *table, int required, int optional, V
 	rest = 1;
 	optional = -1-optional;
     }
-    if (values) {
-	for (j = 0; j < required + optional; j++) {
-	    values[j] = Qundef;
-	}
-    }
     if (required) {
 	for (; i < required; i++) {
 	    VALUE keyword = ID2SYM(table[i]);
 	    if (keyword_hash) {
-		st_data_t val;
-		if (extract_kwarg(keyword, &val)) {
-		    if (values) values[i] = (VALUE)val;
+                if (extract_kwarg(keyword, values[i])) {
 		    continue;
 		}
 	    }
@@ -1898,9 +1937,7 @@ rb_get_kwargs(VALUE keyword_hash, const ID *table, int required, int optional, V
     j = i;
     if (optional && keyword_hash) {
 	for (i = 0; i < optional; i++) {
-	    st_data_t val;
-	    if (extract_kwarg(ID2SYM(table[required+i]), &val)) {
-		if (values) values[required+i] = (VALUE)val;
+            if (extract_kwarg(ID2SYM(table[required+i]), values[required+i])) {
 		j++;
 	    }
 	}
@@ -1910,89 +1947,181 @@ rb_get_kwargs(VALUE keyword_hash, const ID *table, int required, int optional, V
 	    unknown_keyword_error(keyword_hash, table, required+optional);
 	}
     }
+    if (values && !keyword_hash) {
+        for (i = 0; i < required + optional; i++) {
+            values[i] = Qundef;
+        }
+    }
     return j;
 #undef extract_kwarg
 }
 
-#undef rb_scan_args
-int
-rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
-{
-    int i;
-    const char *p = fmt;
-    VALUE *var;
+struct rb_scan_args_t {
+    int argc;
+    const VALUE *argv;
     va_list vargs;
-    int f_var = 0, f_hash = 0, f_block = 0;
-    int n_lead = 0, n_opt = 0, n_trail = 0, n_mand;
-    int argi = 0, last_idx = -1;
-    VALUE hash = Qnil, last_hash = 0;
+    int f_var;
+    int f_hash;
+    int f_block;
+    int n_lead;
+    int n_opt;
+    int n_trail;
+    int n_mand;
+    int argi;
+    int last_idx;
+    VALUE hash;
+    VALUE last_hash;
+    VALUE *tmp_buffer;
+};
+
+static void
+rb_scan_args_parse(int kw_flag, int argc, const VALUE *argv, const char *fmt, struct rb_scan_args_t *arg)
+{
+    const char *p = fmt;
+    VALUE *tmp_buffer = arg->tmp_buffer;
+    int keyword_given = 0;
+    int empty_keyword_given = 0;
+    int last_hash_keyword = 0;
+
+    memset(arg, 0, sizeof(*arg));
+    arg->last_idx = -1;
+    arg->hash = Qnil;
+
+    switch (kw_flag) {
+      case RB_SCAN_ARGS_PASS_CALLED_KEYWORDS:
+        if (!(keyword_given = rb_keyword_given_p())) {
+            empty_keyword_given = rb_empty_keyword_given_p();
+        }
+        break;
+      case RB_SCAN_ARGS_KEYWORDS:
+        keyword_given = 1;
+        break;
+      case RB_SCAN_ARGS_EMPTY_KEYWORDS:
+        empty_keyword_given = 1;
+        break;
+      case RB_SCAN_ARGS_LAST_HASH_KEYWORDS:
+        last_hash_keyword = 1;
+        break;
+    }
 
     if (ISDIGIT(*p)) {
-	n_lead = *p - '0';
+        arg->n_lead = *p - '0';
 	p++;
 	if (ISDIGIT(*p)) {
-	    n_opt = *p - '0';
+            arg->n_opt = *p - '0';
 	    p++;
 	}
     }
     if (*p == '*') {
-	f_var = 1;
+        arg->f_var = 1;
 	p++;
     }
     if (ISDIGIT(*p)) {
-	n_trail = *p - '0';
+        arg->n_trail = *p - '0';
 	p++;
     }
     if (*p == ':') {
-	f_hash = 1;
+        arg->f_hash = 1;
 	p++;
     }
     if (*p == '&') {
-	f_block = 1;
+        arg->f_block = 1;
 	p++;
     }
     if (*p != '\0') {
 	rb_fatal("bad scan arg format: %s", fmt);
     }
-    n_mand = n_lead + n_trail;
-
-    if (argc < n_mand)
-	goto argc_error;
-
-    va_start(vargs, fmt);
+    arg->n_mand = arg->n_lead + arg->n_trail;
 
     /* capture an option hash - phase 1: pop */
-    if (f_hash && n_mand < argc) {
-	VALUE last = argv[argc - 1];
+    /* Ignore final positional hash if empty keywords given */
+    if (argc > 0 && !(arg->f_hash && empty_keyword_given)) {
+        VALUE last = argv[argc - 1];
 
-	if (NIL_P(last)) {
-	    /* nil is taken as an empty option hash only if it is not
-	       ambiguous; i.e. '*' is not specified and arguments are
-	       given more than sufficient */
-	    if (!f_var && n_mand + n_opt < argc)
-		argc--;
-	}
-	else {
-	    hash = rb_check_hash_type(last);
-	    if (!NIL_P(hash)) {
-		VALUE opts = rb_extract_keywords(&hash);
-		if (!(last_hash = hash)) argc--;
-		else last_idx = argc - 1;
-		hash = opts ? opts : Qnil;
-	    }
-	}
+        if (arg->f_hash && arg->n_mand < argc) {
+            if (keyword_given) {
+                if (!RB_TYPE_P(last, T_HASH)) {
+                    rb_warn("Keyword flag set when calling rb_scan_args, but last entry is not a hash");
+                }
+                else {
+                    arg->hash = last;
+                }
+            }
+            else if (NIL_P(last)) {
+                /* For backwards compatibility, nil is taken as an empty
+                   option hash only if it is not ambiguous; i.e. '*' is
+                   not specified and arguments are given more than sufficient.
+                   This will be removed in Ruby 3. */
+                if (!arg->f_var && arg->n_mand + arg->n_opt < argc) {
+                    rb_warn("The last argument is nil, treating as empty keywords");
+                    argc--;
+                }
+            }
+            else {
+                arg->hash = rb_check_hash_type(last);
+            }
+
+            /* Ruby 3: Remove if branch, as it will not attempt to split hashes */
+            if (!NIL_P(arg->hash)) {
+                VALUE opts = rb_extract_keywords(&arg->hash);
+
+                if (!(arg->last_hash = arg->hash)) {
+                    if (!keyword_given && !last_hash_keyword) {
+                        /* Warn if treating positional as keyword, as in Ruby 3,
+                           this will be an error */
+                        rb_warn("Using the last argument as keyword parameters is deprecated");
+                    }
+                    argc--;
+                }
+                else {
+                    /* Warn if splitting either positional hash to keywords or keywords
+                       to positional hash, as in Ruby 3, no splitting will be done */
+                    rb_warn("The last argument is split into positional and keyword parameters");
+                    arg->last_idx = argc - 1;
+                }
+                arg->hash = opts ? opts : Qnil;
+            }
+        }
+        else if (arg->f_hash && keyword_given && arg->n_mand == argc) {
+            /* Warn if treating keywords as positional, as in Ruby 3, this will be an error */
+            rb_warn("Passing the keyword argument as the last hash parameter is deprecated");
+        }
     }
+    if (arg->f_hash && arg->n_mand == argc+1 && empty_keyword_given) {
+        VALUE *ptr = rb_alloc_tmp_buffer2(tmp_buffer, argc+1, sizeof(VALUE));
+        memcpy(ptr, argv, sizeof(VALUE)*argc);
+        ptr[argc] = rb_hash_new();
+        argc++;
+        *(&argv) = ptr;
+        rb_warn("Passing the keyword argument as the last hash parameter is deprecated");
+    }
+
+    arg->argc = argc;
+    arg->argv = argv;
+}
+
+static int
+rb_scan_args_assign(struct rb_scan_args_t *arg, va_list vargs)
+{
+    int argi = 0;
+    int i;
+    VALUE *var;
+
+    if (arg->argc < arg->n_mand) {
+        return 1;
+    }
+
     /* capture leading mandatory arguments */
-    for (i = n_lead; i-- > 0; ) {
+    for (i = arg->n_lead; i-- > 0; ) {
 	var = va_arg(vargs, VALUE *);
-	if (var) *var = (argi == last_idx) ? last_hash : argv[argi];
+        if (var) *var = (argi == arg->last_idx) ? arg->last_hash : arg->argv[argi];
 	argi++;
     }
     /* capture optional arguments */
-    for (i = n_opt; i-- > 0; ) {
+    for (i = arg->n_opt; i-- > 0; ) {
 	var = va_arg(vargs, VALUE *);
-	if (argi < argc - n_trail) {
-	    if (var) *var = (argi == last_idx) ? last_hash : argv[argi];
+        if (argi < arg->argc - arg->n_trail) {
+            if (var) *var = (argi == arg->last_idx) ? arg->last_hash : arg->argv[argi];
 	    argi++;
 	}
 	else {
@@ -2000,15 +2129,15 @@ rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
 	}
     }
     /* capture variable length arguments */
-    if (f_var) {
-	int n_var = argc - argi - n_trail;
+    if (arg->f_var) {
+        int n_var = arg->argc - argi - arg->n_trail;
 
 	var = va_arg(vargs, VALUE *);
 	if (0 < n_var) {
 	    if (var) {
-		int f_last = (last_idx + 1 == argc - n_trail);
-		*var = rb_ary_new4(n_var-f_last, &argv[argi]);
-		if (f_last) rb_ary_push(*var, last_hash);
+                int f_last = (arg->last_idx + 1 == arg->argc - arg->n_trail);
+                *var = rb_ary_new4(n_var - f_last, &arg->argv[argi]);
+                if (f_last) rb_ary_push(*var, arg->last_hash);
 	    }
 	    argi += n_var;
 	}
@@ -2017,18 +2146,18 @@ rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
 	}
     }
     /* capture trailing mandatory arguments */
-    for (i = n_trail; i-- > 0; ) {
+    for (i = arg->n_trail; i-- > 0; ) {
 	var = va_arg(vargs, VALUE *);
-	if (var) *var = (argi == last_idx) ? last_hash : argv[argi];
+        if (var) *var = (argi == arg->last_idx) ? arg->last_hash : arg->argv[argi];
 	argi++;
     }
     /* capture an option hash - phase 2: assignment */
-    if (f_hash) {
+    if (arg->f_hash) {
 	var = va_arg(vargs, VALUE *);
-	if (var) *var = hash;
+        if (var) *var = arg->hash;
     }
     /* capture iterator block */
-    if (f_block) {
+    if (arg->f_block) {
 	var = va_arg(vargs, VALUE *);
 	if (rb_block_given_p()) {
 	    *var = rb_block_proc();
@@ -2037,14 +2166,53 @@ rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
 	    *var = Qnil;
 	}
     }
+
+    if (argi < arg->argc) return 1;
+
+    return 0;
+}
+
+#undef rb_scan_args
+int
+rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
+{
+    int error;
+    va_list vargs;
+    VALUE tmp_buffer = 0;
+    struct rb_scan_args_t arg;
+    arg.tmp_buffer = &tmp_buffer;
+    rb_scan_args_parse(RB_SCAN_ARGS_PASS_CALLED_KEYWORDS, argc, argv, fmt, &arg);
+    va_start(vargs,fmt);
+    error = rb_scan_args_assign(&arg, vargs);
     va_end(vargs);
-
-    if (argi < argc) {
-      argc_error:
-	rb_error_arity(argc, n_mand, f_var ? UNLIMITED_ARGUMENTS : n_mand + n_opt);
+    if (tmp_buffer) {
+        rb_free_tmp_buffer(&tmp_buffer);
     }
+    if (error) {
+        rb_error_arity(arg.argc, arg.n_mand, arg.f_var ? UNLIMITED_ARGUMENTS : arg.n_mand + arg.n_opt);
+    }
+    return arg.argc;
+}
 
-    return argc;
+int
+rb_scan_args_kw(int kw_flag, int argc, const VALUE *argv, const char *fmt, ...)
+{
+    int error;
+    va_list vargs;
+    VALUE tmp_buffer = 0;
+    struct rb_scan_args_t arg;
+    arg.tmp_buffer = &tmp_buffer;
+    rb_scan_args_parse(kw_flag, argc, argv, fmt, &arg);
+    va_start(vargs,fmt);
+    error = rb_scan_args_assign(&arg, vargs);
+    va_end(vargs);
+    if (tmp_buffer) {
+        rb_free_tmp_buffer(&tmp_buffer);
+    }
+    if (error) {
+        rb_error_arity(arg.argc, arg.n_mand, arg.f_var ? UNLIMITED_ARGUMENTS : arg.n_mand + arg.n_opt);
+    }
+    return arg.argc;
 }
 
 int
